"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var types_1 = require("./types");
var events_1 = require("events");
/**
 * Is the controller that can be plugged into a EventManager to store/dump/read the broadcast history
 * of the said EventManager.
 */
var BroadcastsRecorder = /** @class */ (function () {
    function BroadcastsRecorder() {
        this.__eventEmitter = new events_1.EventEmitter();
        /**
         * @private The list of broadcasts that still haven't been dumped
         */
        this.__undumpedBroadcasts = new Map();
        /**
         * The index of the last read message
         */
        this.readOffset = 0;
    }
    Object.defineProperty(BroadcastsRecorder.prototype, "size", {
        /**
         * The total amount of stored undumped broadcasts
         */
        get: function () {
            return this.__undumpedBroadcasts.size;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Reads an entry from the records
     * Beware that unbound methods or getters/setters are not saved on the records.
     * @param index the index of the record to be read
     */
    BroadcastsRecorder.prototype.read = function (index) {
        var inputReadIndex = (index !== null && index !== void 0 ? index : this.nextOffset);
        var selectedReadIndex = inputReadIndex >= this.size
            ? this.size
            : inputReadIndex;
        var result = this.__undumpedBroadcasts.get(selectedReadIndex);
        if (index === undefined && result !== undefined && inputReadIndex <= this.size) {
            this.readOffset += 1;
        }
        this.__eventEmitter.emit(types_1.BroadcastsRecorderEvents.READ, [result]);
        return result;
    };
    Object.defineProperty(BroadcastsRecorder.prototype, "nextOffset", {
        /**
         * The next offset that the method .read() will use
         */
        get: function () {
            return this.readOffset + 1;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Writes the state to the undumped broadcast list
     * @param procedureName
     * @param contextState
     * @param data Optional
     * @WARNING Writing directly to this will make this instance record a change
     * that did not happen!
     */
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    BroadcastsRecorder.prototype.writeState = function (procedureName, contextState, data) {
        var newEntry = {
            contextState: Object.assign({}, contextState),
            data: data,
            procedureName: procedureName,
        };
        this.__undumpedBroadcasts.set(this.size + 1, newEntry);
        this.__eventEmitter.emit(types_1.BroadcastsRecorderEvents.WRITE, [newEntry]);
        return;
    };
    /**
     * Reads the list of recorded broadcasts and clears it.
     * Beware that unbound methods or getters/setters are not saved on the records.
     */
    BroadcastsRecorder.prototype.dump = function () {
        var result = [];
        this.__undumpedBroadcasts.forEach(function (value) {
            result.push(value);
        });
        this.__undumpedBroadcasts.clear();
        this.readOffset = 0;
        this.__eventEmitter.emit(types_1.BroadcastsRecorderEvents.DUMP, result);
        return result;
    };
    /**
     * Listens to an event that is emitted by this BroadcastsRecorder instance
     * @param event The event to listen to.
     * @param callback The callback to be executed when the event is emitted.
     */
    BroadcastsRecorder.prototype.on = function (event, callback) {
        this.__eventEmitter.on(event, callback);
        return this;
    };
    /**
     * Removes a listener from this BroadcastsRecorder instance
     * @param event The event to remove listener from.
     * @param callback The listener callback.
     */
    BroadcastsRecorder.prototype.off = function (event, callback) {
        this.__eventEmitter.removeListener(event, callback);
        return this;
    };
    return BroadcastsRecorder;
}());
exports.BroadcastsRecorder = BroadcastsRecorder;
