"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Context defines a group of domain information. It available to a Procedure when
 * it gets executed. A context may have any Bibrbable instance signed on it.
 */
// eslint-disable-next-line @typescript-eslint/no-explicit-any
var Context = /** @class */ (function () {
    /**
     * @param identifier The value to set the local identifier of the context
     */
    function Context(contextOtions) {
        var _a;
        /**
         * @readonly @private Map of Birbables registered in the context
         */
        this.__birbables = new Map();
        this.__contextState = {};
        this.__identifier = contextOtions.identifier;
        this.__errorHandler = (_a = contextOtions.errorHandler) === null || _a === void 0 ? void 0 : _a.bind(this);
    }
    Object.defineProperty(Context.prototype, "contextState", {
        /**
         * The current state data of the context
         */
        get: function () {
            return this.__contextState;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Sets a state to the context
     * @param information The object information to be set
     */
    Context.prototype.setContextState = function (information) {
        Object.assign(this.__contextState, information);
        return this;
    };
    Object.defineProperty(Context.prototype, "identifier", {
        /**
         * @returns identifier of the context entity
         */
        get: function () {
            return this.__identifier;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Context.prototype, "errorHandler", {
        set: function (handler) {
            this.__errorHandler = handler;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Triggers the execution of a signed `Birbable`.
     * @param name Identifier of what needs to be triggered in this context
     * @param descriptable Additional Information to be sent to the Birbable
     * @param errorHandler A function to catch errors in case of unexpected promise Rejections
     */
    Context.prototype.trigger = function (options, descriptable) {
        var foundBirbable = this.__birbables.get(options.birbable);
        if (foundBirbable === undefined) {
            return this;
        }
        this.__unmount(foundBirbable);
        var execution = foundBirbable.execute(this, descriptable);
        if (execution instanceof Promise) {
            execution.catch(this.__activeErrorHandler(options.errorHandler));
        }
        return this;
    };
    Context.prototype.__activeErrorHandler = function (errorHandler) {
        var _this = this;
        var _a, _b;
        var defaultHandler = function (error) {
            console.log("The context \"" + _this.constructor.name + "\" had an Unhandled Promise Rejection.");
            throw error;
        };
        return (_b = ((_a = this.__errorHandler) !== null && _a !== void 0 ? _a : errorHandler)) !== null && _b !== void 0 ? _b : defaultHandler;
    };
    /**
     * Removes a birbable which is about to be triggered
     * @private
     * @param birbable
     */
    Context.prototype.__unmount = function (birbable) {
        if (birbable.belongsToGroup) {
            return this.__clearGroup(birbable.group);
        }
        if (birbable.lifetime === 'SINGLE') {
            this.__birbables.delete(birbable.constructor.name);
        }
    };
    /**
     * Removes a whole group of birbables
     * @private
     * @param group a symbol marking the group to be removed
     */
    Context.prototype.__clearGroup = function (group) {
        var _this = this;
        var groupEntities = [];
        this.__birbables.forEach(function (birbable) {
            if (birbable.group === group) {
                groupEntities.push(birbable);
            }
        });
        groupEntities.forEach(function (groupedBirbable) {
            _this.__birbables.delete(groupedBirbable.constructor.name);
        });
    };
    /**
     * Signs a Birbable in your context. Only signed Birbables can be triggered
     * @param birbable
     */
    Context.prototype.sign = function (birbable) {
        this.__birbables.set(birbable.constructor.name, birbable);
        // this.__addToListener(birbable);
        return this;
    };
    /**
     * Usigns a Birbable from the context.
     * @param birbable Any Birbable that was signed before in this context
     */
    Context.prototype.unsign = function (birbable) {
        var foundBirbable = this.__birbables.get(birbable.constructor.name);
        if (foundBirbable !== undefined) {
            this.__birbables.delete(birbable.constructor.name);
            // this.__emitter.removeAllListeners(birbable.constructor.name);
        }
        return this;
    };
    return Context;
}());
exports.Context = Context;
